import json
import os
import shutil
import tempfile
import unittest
from pathlib import Path
from retry import retry

from filoc import Filoc
from filoc.rawfiloc import RawFiloc
import time


def touch_file(file_path):
    os.makedirs(os.path.dirname(file_path))
    Path(file_path).touch()


# noinspection DuplicatedCode
class TestFiloc(unittest.TestCase):
    """
    TODO: 
    - Test cache behavior on delete of files (should currently fail --> TODO DEV Feature)    
    """


    def setUp(self):
        self.maxDiff = None
        self.test_dir = tempfile.mkdtemp().replace('\\', '/')
        self.path_fmt = self.test_dir + r'/simid={simid:d}/epid={epid:d}/hyperparameters.json'

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_read_contents(self):
        wloc = Filoc(self.path_fmt, writable=True)
        with wloc.open({"simid": 1, "epid": 10}, "w") as f: json.dump({'a': 100}, f)
        with wloc.open({"simid": 1, "epid": 20}, "w") as f: json.dump({'a': 200}, f)
        with wloc.open({"simid": 2, "epid": 10}, "w") as f: json.dump({'a': 300}, f)
        with wloc.open({"simid": 2, "epid": 20}, "w") as f: json.dump({'a': 400}, f)

        loc = Filoc(self.path_fmt)
        p = loc.read_contents({'epid': 10})
        self.assertEqual(len(p), 2)
        self.assertEqual('[{"a": 100, "epid": 10, "simid": 1}, {"a": 300, "epid": 10, "simid": 2}]',
                         json.dumps(p, sort_keys=True))

        loc = Filoc(self.path_fmt, content_reader=reporter_1param)
        p = loc.read_contents({'epid': 10})
        self.assertEqual(len(p), 2)
        self.assertEqual('[{"aaa": 100, "epid": 10, "simid": 1}, {"aaa": 300, "epid": 10, "simid": 2}]',
                         json.dumps(p, sort_keys=True))

        loc = Filoc(self.path_fmt, content_reader=reporter_2param)
        p = loc.read_contents({'epid': 10})
        self.assertEqual(len(p), 2)
        self.assertEqual('[{"aaa": 100, "b": 1, "epid": 10, "simid": 1}, {"aaa": 300, "b": 2, "epid": 10, "simid": 2}]',
                         json.dumps(p, sort_keys=True))

        # change file
        with wloc.open({"simid": 2, "epid": 10}, "w") as f: 
            json.dump({'a': 333}, f)
        
        loc = Filoc(self.path_fmt, content_reader=reporter_1param)
        
        @retry(tries=20, delay=2)
        def act_assert():
            p = loc.read_contents({'epid': 10})
            self.assertEqual(len(p), 2)
            self.assertEqual('[{"aaa": 100, "epid": 10, "simid": 1}, {"aaa": 333, "epid": 10, "simid": 2}]', json.dumps(p, sort_keys=True))

        act_assert()


    def test_read_contents_with_cache(self):
        print("write files")
        wloc = Filoc(self.path_fmt, writable=True)
        with wloc.open({"simid": 1, "epid": 10}, "w") as f: json.dump({'a': 100}, f)
        with wloc.open({"simid": 1, "epid": 20}, "w") as f: json.dump({'a': 200}, f)
        with wloc.open({"simid": 2, "epid": 10}, "w") as f: json.dump({'a': 300}, f)
        with wloc.open({"simid": 2, "epid": 20}, "w") as f: json.dump({'a': 400}, f)

        loc = Filoc(self.path_fmt, cache_locpath='.cache')
        print("read_contents 1")
        p = loc.read_contents({'epid': 10})
        self.assertEqual(len(p), 2)
        self.assertEqual('[{"a": 100, "epid": 10, "simid": 1}, {"a": 300, "epid": 10, "simid": 2}]',
                         json.dumps(p, sort_keys=True))

        time.sleep(0.1) # small delay between the two writings to ensure that the file gets an new timestamp

        print("change one file")
        # change to file triggers cache refresh
        with wloc.open({"simid": 2, "epid": 10}, "w") as f: 
            json.dump({'a': 333}, f)
            # f.flush()
            # os.fsync(f.fileno())

        # @retry(tries=10, delay=1)
        # def act_assert():
        print("re read_contents 2")
        p = loc.read_contents({'epid': 10})
        self.assertEqual(len(p), 2)
        self.assertEqual('[{"a": 100, "epid": 10, "simid": 1}, {"a": 333, "epid": 10, "simid": 2}]',
                        json.dumps(p, sort_keys=True))

        # act_assert()

        # Trick to test: signature change does not take effect, because of cache
        loc = Filoc(self.path_fmt, content_reader=reporter_1param, cache_locpath='.cache')
        print("re read_contents 3")
        p = loc.read_contents({'epid': 10})
        self.assertEqual(len(p), 2)
        self.assertEqual('[{"a": 100, "epid": 10, "simid": 1}, {"a": 333, "epid": 10, "simid": 2}]',
                         json.dumps(p, sort_keys=True))

    def test_write_contents(self):
        wloc = Filoc(self.path_fmt, writable=True)
        wloc.write_contents([
            {"simid": 1, "epid": 10, 'a': 100},
            {"simid": 1, "epid": 20, 'a': 200},
            {"simid": 2, "epid": 10, 'a': 300},
            {"simid": 2, "epid": 20, 'a': 400},
        ])

        wloc = Filoc(self.path_fmt)
        with wloc.open({"simid": 1, "epid": 10}) as f: c1 = json.load(f)
        with wloc.open({"simid": 1, "epid": 20}) as f: c2 = json.load(f)
        with wloc.open({"simid": 2, "epid": 10}) as f: c3 = json.load(f)
        with wloc.open({"simid": 2, "epid": 20}) as f: c4 = json.load(f)

        self.assertEqual('{"a": 100}', json.dumps(c1, sort_keys=True))
        self.assertEqual('{"a": 200}', json.dumps(c2, sort_keys=True))
        self.assertEqual('{"a": 300}', json.dumps(c3, sort_keys=True))
        self.assertEqual('{"a": 400}', json.dumps(c4, sort_keys=True))


def reporter_1param(f):
    content = json.load(f)
    return {'aaa': content['a']}


def reporter_2param(f, properties):
    content = json.load(f)
    return {
        'aaa': content['a'],
        'b': properties['simid']
    }


if __name__ == '__main__':
    unittest.main()



from .rawfiloc import RawFiloc
from .filoc import Filoc
from .multiloc import MultilocBase, Multiloc, PandasMultiloc

__version__ = '0.0.6'
__all__     = [
    'RawFiloc',
    'Filoc',
    'MultilocBase',
    'Multiloc',
    'PandasMultiloc',
]


import json
import logging
import os
import pickle
from collections import OrderedDict
from functools import partial
from inspect import getfullargspec
from io import UnsupportedOperation
from typing import Dict, Any, List, Union, Callable, Optional, Literal, Tuple, IO

from frozendict import frozendict

from .rawfiloc import RawFiloc, mix_dicts

log = logging.getLogger('filoc')

# -----
# Types
# -----
PresetFiLocFileTypes = Literal[None, 'json', 'yaml', 'pickle', 'csv']

FileReader = Union[
    Callable[[IO], Dict[str, Any]],
    Callable[[IO, Dict[str, Any]], Dict[str, Any]]
]

FileWriter = Union[
    Callable[[Dict[str, Any], IO], None],
    Callable[[Dict[str, Any], Dict[str, Any], IO], None]
]

_default_cache_reader = pickle.load
_default_cache_writer = pickle.dump


# -------
# Helpers
# -------
def _get_content_reader_writer(file_type : PresetFiLocFileTypes):
    if file_type == 'json':
        import json
        return json.load, partial(json.dump, indent=2, sort_keys=True)
    elif file_type == 'yaml':
        # noinspection PyPackageRequirements
        import yaml
        return yaml.load, yaml.dump
    elif file_type == 'csv':
        import csv
        return csv.DictReader, csv.DictWriter
    elif file_type == 'pickle':
        import pickle
        return pickle.load, pickle.dump


# -------------------
# Main Class Filoc
# -------------------
class Filoc(RawFiloc):
    # noinspection PyDefaultArgument
    def __init__(
            self,
            locpath                : str,
            file_type              : PresetFiLocFileTypes = 'json',
            writable               : bool                 = False,
            content_reader         : Optional[FileReader] = None,
            content_writer         : Optional[FileWriter] = None,
            content_reader_options : Dict[str, Any]       = {'mode': 'r'},
            content_writer_options : Dict[str, Any]       = {'mode': 'w'},
            cache_locpath          : str                  = None,
            timestamp_col          : str                  = None,
    ):
        """
        if cache_locpath is relative, then it will be relative to result_locpath
        """
        super().__init__(locpath, writable=writable)

        # reader and writer
        self.file_type = file_type
        default_reader, default_writer = (None, None)
        if file_type and (content_reader is None or content_writer is None):
            default_reader, default_writer = _get_content_reader_writer(file_type)
        self.content_reader         = content_reader if content_reader else default_reader if file_type else None  # type:Optional[FileReader]
        self.content_writer         = content_writer if content_writer else default_writer if file_type else None  # type:Optional[FileWriter]
        self.content_reader_options = content_reader_options  # type:Dict[str, Any]
        self.content_writer_options = content_writer_options  # type:Dict[str, Any]

        # cache loc
        self.cache_loc = None
        if cache_locpath is not None:
            if not os.path.isabs(cache_locpath):
                cache_locpath = self.root_folder + '/' + cache_locpath
            self.cache_loc = RawFiloc(cache_locpath, writable=True)
        self.timestamp_col = timestamp_col
        # cache_timestamp_col
        if self.cache_loc is None:
            self.cache_timestamp_col = None
        elif timestamp_col is None:
            self.cache_timestamp_col = 'timestamp'
        else:
            self.cache_timestamp_col = timestamp_col

    def invalidate_cache(self, path_props : Optional[Dict[str, Any]] = None, **path_props_kwargs):
        if self.cache_loc is None:
            return
        path_props = mix_dicts(path_props, path_props_kwargs)
        self.cache_loc.delete(path_props)

    def read_content(self, path_props : Optional[Dict[str, Any]] = None, **path_props_kwargs) -> List[Dict[str, Any]]:
        path_props = mix_dicts(path_props, path_props_kwargs)
        path = self.get_path(path_props)  # validates, that pat_props points to a single file
        contents = self.read_contents(path_props)
        if len(contents) == 0:
            raise ValueError(f'No keyvalues found for path props {path_props} (path={path})')
        return contents[0]

    def read_contents(self, path_props : Optional[Dict[str, Any]] = None, **path_props_kwargs) -> List[Dict[str, Any]]:
        path_props = mix_dicts(path_props, path_props_kwargs)
        result = []
        for (props, values) in self._read_file_content_by_path_props(path_props).items():
            entry = OrderedDict(values)
            entry.update(props)  # path properties have precedence!
            result.append(entry)
        return result

    def _read_file_content_by_path_props(self, path_props : Optional[Dict[str, Any]] = None, **path_props_kwargs) -> Dict[Dict[str, Any], Dict[str, Any]]:
        path_props = mix_dicts(path_props, path_props_kwargs)
        result = OrderedDict()

        cache_props_and_cache = None  # type:Optional[Tuple[Dict[str, Any]], Dict[Dict[str, Any], Dict[str, Any]]]
        paths_and_path_props  = self.find_paths_and_path_props(path_props)
        log.info(f'Found {len(paths_and_path_props)} files to read in locpath {self.locpath} with path_props {json.dumps(path_props)}')
        for (f_path, f_props) in paths_and_path_props:
            f_props_hashable = frozendict(f_props.items())

            try:
                f_timestamp = self.fs.modified(f_path)
            except FileNotFoundError:
                f_timestamp = None
                
            # renew cache, on cache file change
            if self.cache_loc:
                f_cache_path  = self.cache_loc.get_path(f_props)
                f_cache_props = self.cache_loc.get_path_properties(f_cache_path)

                if cache_props_and_cache is not None and cache_props_and_cache[0] == f_cache_props:
                    pass  # do nothing, keep this cache file opened
                else:
                    if cache_props_and_cache is not None and cache_props_and_cache[0] != f_cache_props:
                        # flush previous cache, if exists
                        if cache_props_and_cache[0]:
                            with self.cache_loc.open(cache_props_and_cache[0], 'wb') as f:
                                _default_cache_writer(cache_props_and_cache[1], f)

                    # now prepare new cache
                    if self.cache_loc.exists(f_cache_props):
                        with self.cache_loc.open(f_cache_props, 'rb') as f:
                            cache_props_and_cache = (f_cache_props, _default_cache_reader(f))
                    else:
                        cache_props_and_cache = (f_cache_props, OrderedDict())

            # check whether cache entry is still valid
            if self.cache_loc:
                if f_props_hashable in cache_props_and_cache[1]:
                    cached_entry = cache_props_and_cache[1][f_props_hashable]  # type: Dict[str, Any]
                    cached_entry_timestamp = cached_entry[self.cache_timestamp_col]
                    if cached_entry_timestamp == f_timestamp:
                        print(f'File analysis cached: {f_path}')
                        log.info(f'File analysis cached: {f_path}')
                        result[f_props_hashable] = cached_entry.copy()  # copy from cache
                        continue
                    else:
                        print(f'Cache out of date for {f_path}')
                        log.info(f'Cache out of date for {f_path}')

            # cache is not valid: read file!
            f_keyvalues = OrderedDict()
            # -> timestamps
            if self.cache_timestamp_col:
                f_keyvalues[self.cache_timestamp_col] = f_timestamp
            if self.timestamp_col:
                f_keyvalues[self.timestamp_col] = f_timestamp
            # -> custom report
            if self.content_reader is not None:
                f_keyvalues.update(self._read_single_file(f_path, f_props))

            # add to result and cache
            result[f_props_hashable] = f_keyvalues
            if self.cache_loc:
                cache_props_and_cache[1][f_props_hashable] = f_keyvalues.copy()  # copy to cache

        # flush last used cache
        if cache_props_and_cache:
            with self.cache_loc.open(cache_props_and_cache[0], 'wb') as f:
                _default_cache_writer(cache_props_and_cache[1], f)

        if self.cache_timestamp_col != self.timestamp_col:
            for f_keyvalues in result.values():
                if self.cache_timestamp_col in f_keyvalues:
                    del f_keyvalues[self.cache_timestamp_col]

        return result

    def write_content(self, keyvalues : Dict[str, Any], dry_run=False):
        if not self.writable:
            raise UnsupportedOperation('this filoc is not writable. Set writable flag to True to enable writing')
        self.write_contents([keyvalues], dry_run=dry_run)

    def write_contents(self, keyvalues_list : List[Dict[str, Any]], dry_run=False):
        if not self.writable:
            raise UnsupportedOperation('this filoc is not writable. Set writable flag to True to enable writing')
        recorded_row_id_by_path_props    = {}
        recorded_keyvalues_by_path_props = {}
        for row_id, keyvalues in enumerate(keyvalues_list):
            f_path_props, timestamp, cache_timestamp, f_keyvalues = self._split_keyvalues(keyvalues)
            f_path_props_hashable = frozendict(f_path_props)

            # avoid saving file if no changes
            if self.exists(f_path_props):
                old_keyvalues = self.read_content(f_path_props)
                if old_keyvalues == keyvalues:
                    log.info(f'Row {row_id}: No difference detected: do nothing')
                    continue

            log.info(f'Row {row_id}: Difference detected: update values')

            # validate multiple changes consistency
            if f_path_props_hashable in recorded_keyvalues_by_path_props:
                recorded_keyvalues = recorded_keyvalues_by_path_props[f_path_props_hashable]
                if recorded_keyvalues != f_keyvalues:
                    other_row_id = recorded_row_id_by_path_props[f_path_props_hashable]
                    raise ValueError(f'Rows {row_id} and {other_row_id} need to be saved to the same file, but have different keyvalues:\n{row_id}: {f_keyvalues}\n{other_row_id}: {recorded_keyvalues}')
                else:
                    continue  # do nothing, as change has already been recorded

            # record change to perform and notice row_id (last is for error message)
            recorded_row_id_by_path_props[f_path_props_hashable]    = row_id
            recorded_keyvalues_by_path_props[f_path_props_hashable] = f_keyvalues

        dry_run_log_prefix = '(dry_run) ' if dry_run else ''
        for f_path_props, keyvalues in recorded_keyvalues_by_path_props.items():
            self.invalidate_cache(f_path_props)
            path = self.get_path(f_path_props)
            with self.open(f_path_props, **self.content_writer_options) as f:
                log.info(f'{dry_run_log_prefix}Saving to {path}: {json.dumps(keyvalues)}')
                if dry_run:
                    continue
                self.content_writer(keyvalues, f)
                log.info(f'{dry_run_log_prefix}Saved {path}: {json.dumps(keyvalues)}')

    def _split_keyvalues(self, keyvalues):
        path_props      = {}
        timestamp       = None
        cache_timestamp = None
        file_keyvalues  = OrderedDict()
        for (k, v) in keyvalues.items():
            if k in self.path_props:
                path_props[k] = v
            elif k == self.timestamp_col:
                timestamp = keyvalues[k]
            elif k == self.cache_timestamp_col:
                cache_timestamp = keyvalues[k]
            else:
                file_keyvalues[k] = v
        return path_props, timestamp, cache_timestamp, file_keyvalues

    def _read_single_file(self, f_path, f_props):
        len_params = len(getfullargspec(self.content_reader).args)
        if len_params == 1:
            log.info(f'Analyzing file {f_path}')
            with self.fs.open(f_path, **self.content_reader_options) as f:
                custom_report = self.content_reader(f)
            log.info(f'Analyzed file {f_path}')
        elif len_params == 2:
            log.info(f'Analyzing file {f_path}')
            with self.fs.open(f_path, **self.content_reader_options) as f:
                custom_report = self.content_reader(f, f_props)
            log.info(f'Analyzed file {f_path}')
        else:
            raise ValueError(f'Provided content_reader has {len_params} parameters, allowed signature: f(path) or f(path, path_props)')
        return custom_report

    def _write_single_file(self, f_keyvalues : Dict[str, Any], f_path : str, f_path_props : Dict[str, Any]):
        len_params = len(getfullargspec(self.content_writer).args)
        if len_params == 1:
            log.info(f'Analyzing file {f_path}')
            with self.fs.open(f_path, **self.content_reader_options) as f:
                custom_report = self.content_writer(f_keyvalues, f)
            log.info(f'Analyzed file {f_path}')
        elif len_params == 2:
            log.info(f'Analyzing file {f_path}')
            with self.fs.open(f_path, **self.content_reader_options) as f:
                custom_report = self.content_writer(f_keyvalues, f_path_props, f)
            log.info(f'Analyzed file {f_path}')
        else:
            raise ValueError(f'Provided content_writer requires {len_params} positional parameters, allowed signature: f(path) or f(path, path_props)')
        return custom_report



import logging
from collections import OrderedDict
from typing import Dict, Any, List, Optional, Set, Tuple, Callable, TypeVar, Generic, get_args

from orderedset import OrderedSet

from .filoc import Filoc
from .rawfiloc import mix_dicts

log = logging.getLogger('multiloc')

TContent = TypeVar('TContent')


class MultilocBase(Generic[TContent]):
    def __init__(
            self,
            filoc_by_name    : Dict[str, Filoc],
            make_content_fn  : Callable[  [  List[Dict[Tuple[str, str], Any]], Tuple[str, str]  ], TContent  ],
            parse_content_fn : Callable[  [TContent], List[Dict[Tuple[str,str], Any]]],
            path_level_name  : str,
    ):
        assert isinstance(filoc_by_name, dict)

        self.filoc_by_name    = filoc_by_name
        self.path_level_name  = path_level_name
        self.make_content_fn  =  make_content_fn
        self.parse_content_fn = parse_content_fn

    def read_contents(self, path_props : Optional[Dict[str, Any]] = None, **path_props_kwargs) -> TContent:
        path_props = mix_dicts(path_props, path_props_kwargs)
        contents, keys = self._read_raw_contents_and_keys(path_props)
        return self.make_content_fn(contents, keys)

    def write_contents(self, content : TContent, dry_run=False):
        raw_content = self.parse_content_fn(content)
        self._write_raw_contents(raw_content, dry_run=dry_run)

    def _read_raw_contents_and_keys(self, path_props : Optional[Dict[str, Any]] = None) -> Tuple[List[Dict[Tuple[str, str], Any]], Tuple[str, str]]:
        # collect
        all_path_props_combinations_set = set()             # type:Set[Dict[str, Any]]
        all_path_props_combinations_in_order = []           # type:List[Dict[str, Any]]
        keyvalues_by_path_props_by_filoc_name = OrderedDict()  # type:OrderedDict[str, Dict[Dict[str, Any], Dict[str, Any]]]
        for filoc_name, filoc in self.filoc_by_name.items():
            # noinspection PyProtectedMember
            keyvalues_by_path_props = filoc._read_file_content_by_path_props(path_props)
            keyvalues_by_path_props_by_filoc_name[filoc_name] = keyvalues_by_path_props
            for props in keyvalues_by_path_props:
                if props not in all_path_props_combinations_set:
                    all_path_props_combinations_set.add(props)
                    all_path_props_combinations_in_order.append(props)

        # outer join
        result_keys = OrderedSet()
        result = []
        for path_props in all_path_props_combinations_in_order:
            result_row = OrderedDict()
            result.append(result_row)

            # add path_props
            for property_name, property_value in path_props.items():
                hierarchical_key = (self.path_level_name, property_name)
                result_row[hierarchical_key] = property_value
                result_keys.add(hierarchical_key)

            # add all key values for each filoc
            for filoc_name, keyvalues_by_path_props in keyvalues_by_path_props_by_filoc_name.items():
                keyvalues = keyvalues_by_path_props.get(path_props, None)
                if keyvalues:
                    for key, value in keyvalues.items():
                        hierarchical_key = (filoc_name, key)
                        result_row[hierarchical_key] = value
                        result_keys.add(hierarchical_key)
        return result, list(result_keys)

    def _write_raw_contents(self, raw_content : List[Dict[Tuple[str, str], Any]], dry_run=False):
        # split keyvalues_list into filoc buckets
        # ... first create empty placeholder vor key values
        path_props_list = [OrderedDict() for _ in range(len(raw_content))]
        keyvalues_list_by_filoc_name = {}
        for filoc_name, filoc in self.filoc_by_name.items():
            if filoc.content_writer is None:
                continue  # don't write non writable filocs
            filoc_keyvalues_list = []
            keyvalues_list_by_filoc_name[filoc_name] = filoc_keyvalues_list
            for i in range(len(raw_content)):
                filoc_keyvalues_list.append(OrderedDict())

        # ... then fill
        for row_id, keyvalues in enumerate(raw_content):
            for (filoc_name, prop_name), value in keyvalues.items():
                if filoc_name == self.path_level_name:
                    row_path_props = path_props_list[row_id]
                    row_path_props[prop_name] = value
                elif filoc_name in keyvalues_list_by_filoc_name:
                    filoc_row_keyvalues = keyvalues_list_by_filoc_name[filoc_name][row_id]
                    filoc_row_keyvalues[prop_name] = value
                else:
                    log.warning(f'write_contents: Skip unknown Filoc name {filoc_name} for hierarchy {key}')

        # delegate writing to
        for filoc_name, filoc_keyvalues_list in keyvalues_list_by_filoc_name.items():
            # merge path_props to filoc_keyvalues for each row
            for row_id, filoc_keyvalues in enumerate(filoc_keyvalues_list):
                filoc_keyvalues.update(path_props_list[row_id])
            filoc = self.filoc_by_name[filoc_name]
            filoc.write_contents(filoc_keyvalues_list, dry_run=dry_run)


def make_json_content_factory(separator : str) -> Callable[  [  List[Dict[Tuple[str, str], Any]], Tuple[str, str]  ], List[Dict[str, Any]]  ]:
    # noinspection PyUnusedLocal
    def make_default_json_content(raw_contents : List[Dict[Tuple[str,str], Any]], keys : List[Tuple[str,str]]) -> List[Dict[str, Any]]:
        return [{separator.join(k): v for (k, v) in row.items()} for row in raw_contents]
    return make_default_json_content


def parse_json_content_factory(separator : Optional[str] = None) -> Callable[  [  List[Dict[str, Any]]  ], List[Dict[Tuple[str,str], Any]]  ]:
    def parser_default_json_content(contents : List[Dict[str, Any]]) -> List[Dict[Tuple[str,str], Any]]:
        return [{tuple(k.split(separator)): v for (k, v) in row.items()} for row in contents]
    return parser_default_json_content


class Multiloc(MultilocBase[List[Dict[Tuple[str, str], Any]]]):
    def __init__(
            self,
            filoc_by_name: Dict[str, Filoc],
            level_separator: str = ".",
            path_level_name: str = 'path'
    ):
        super().__init__(
            filoc_by_name,
            make_json_content_factory(level_separator),
            parse_json_content_factory(level_separator),
            path_level_name
        )


try:
    import pandas  # pandas is not a required dependency of filoc
    from pandas import DataFrame

    def make_pandas_content_factory(separator : Optional[str] = None) -> Callable[  [  List[Dict[Tuple[str, str], Any]], Tuple[str, str]  ], DataFrame]:
        def make_content_pandas(raw_contents : List[Dict[Tuple[str,str], Any]], keys : List[Tuple[str,str]]) -> DataFrame:
            if separator:
                keys = [separator.join(k) for k in keys]
                raw_contents = [{separator.join(k): v for (k, v) in row.items()} for row in raw_contents]
            return pandas.DataFrame(raw_contents, columns=keys)
        return make_content_pandas


    def parse_pandas_content_factory(separator : Optional[str] = None) -> Callable[[DataFrame], List[Dict[Tuple[str,str], Any]]]:
        def parser_content_pandas(contents : DataFrame) -> List[Dict[Tuple[str,str], Any]]:
            raw_contents = contents.to_dict(orient='records')
            if separator:
                raw_contents = [{tuple(k.split(separator)): v for (k, v) in row.items()} for row in raw_contents]
            return raw_contents
        return parser_content_pandas


    class PandasMultiloc(MultilocBase[DataFrame]):
        def __init__(
                self,
                filoc_by_name: Dict[str, Filoc],
                level_separator: str = ".",
                path_level_name: str = 'path'
        ):
            super().__init__(
                filoc_by_name,
                make_pandas_content_factory(level_separator),
                parse_pandas_content_factory(level_separator),
                path_level_name
            )

except ModuleNotFoundError:
    class PandasMultiloc(MultilocBase):
        # noinspection PyMissingConstructor
        def __init__(self, *args, **kwargs):
            raise ModuleNotFoundError("pandas not available: pip install pandas / conda install pandas")


